package org.verohallinto.apitamoclient.sec;

import org.apache.commons.lang.exception.ExceptionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.verohallinto.apitamoclient.util.Utils;

import javax.net.ssl.SSLEngine;
import javax.net.ssl.X509ExtendedKeyManager;
import java.io.FileNotFoundException;
import java.net.Socket;
import java.security.KeyStore;
import java.security.Principal;
import java.security.PrivateKey;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;

/**
 * AliasKeyManager which allows to always select predefined client certificate.
 */
public class AliasKeyManager extends X509ExtendedKeyManager {

    private static final Logger log = LoggerFactory.getLogger(AliasKeyManager.class);
    private KeyStore ks;
    private String alias;
    private String aliasPwd;

    /**
     * Constructor.
     * </p>
     *
     * @param ks       <code>KeyStore</code> KeyStore containing client's certificate.
     * @param alias    <code>String</code> certificate alias name.
     * @param aliasPwd <code>String</code> alias password.
     */
    public AliasKeyManager(final KeyStore ks, final String alias, final String aliasPwd) {

        if (ks == null) {
            throw new IllegalArgumentException("The KeyStore cannot be null.");
        }

        if (Utils.isEmpty(alias)) {
            throw new IllegalArgumentException("The KeyStore alias cannot be empty.");
        }

        if (Utils.isEmpty(aliasPwd)) {
            throw new IllegalArgumentException("The KeyStore alias password cannot be empty.");
        }

        this.ks = ks;
        this.alias = alias;
        this.aliasPwd = aliasPwd;
    }

    /**
     * Returns always the selected key alias.
     * </p>
     *
     * @param s          <code>String</code>
     * @param principals <code>Principal</code> principals.
     * @return Key alias as an array
     */
    @Override
    public String[] getClientAliases(final String s, final Principal[] principals) {

        if (log.isDebugEnabled()) {
            log.debug("getClientAliases() - Alias=" + this.alias);
        }

        return new String[]{this.alias};
    }

    /**
     * Returns always the selected key alias.
     * </p>
     *
     * @param ss         <code>String</code>
     * @param principals <code>Principal</code> principals.
     * @param socket     <code>Socket</code> socket.
     * @return Key alias as a String.
     */
    @Override
    public String chooseClientAlias(final String[] ss, final Principal[] principals, final Socket socket) {

        if (log.isDebugEnabled()) {
            log.debug("chooseClientAlias() - Alias=" + this.alias);
        }

        return this.alias;
    }

    /**
     * Falls back to the original KeyManager.
     * </p>
     *
     * @param s          <code>String</code>
     * @param principals <code>Principal</code> principals.
     * @return Key alias as an array.
     */
    @Override
    public String[] getServerAliases(final String s, final Principal[] principals) {

        if (log.isDebugEnabled()) {
            log.debug("getServerAliases() - Alias=" + this.alias);
        }

        return new String[]{this.alias};
    }

    /**
     * Falls back to the original KeyManager.
     * </p>
     *
     * @param s          <code>String</code>
     * @param principals <code>Principal</code> principals.
     * @param socket     <code>Socket</code> socket.
     * @return Key alias as a String.
     */
    @Override
    public String chooseServerAlias(final String s, final Principal[] principals, final Socket socket) {

        if (log.isDebugEnabled()) {
            log.debug("chooseServerAlias() - Alias=" + this.alias);
        }

        return this.alias;
    }

    /**
     * Falls back to the original KeyManager.
     * </p>
     *
     * @param s <code>String</code>
     * @return Certificate chain.
     */
    @Override
    public X509Certificate[] getCertificateChain(final String s) {

        if (log.isDebugEnabled()) {
            log.debug("getCertificateChain() - begin. Alias=" + this.alias);
        }

        X509Certificate[] x509Certificates = null;

        try {
            Certificate[] certificates = this.ks.getCertificateChain(this.alias);

            if (certificates == null) {
                throw new FileNotFoundException("getCertificateChain() - No currCertificate found for Alias: " + this.alias);
            }

            x509Certificates = new X509Certificate[certificates.length];
            System.arraycopy(certificates, 0, x509Certificates, 0, certificates.length);
        } catch (Exception e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
        }

        if (log.isDebugEnabled()) {
            log.debug("getCertificateChain() - end. x509Certificates:");

            if (x509Certificates != null) {
                for (X509Certificate crt : x509Certificates) {
                    log.debug("\t" + crt.getSubjectDN().getName());
                }
            }
        }

        return x509Certificates;
    }

    /**
     * @param alias <code>String</code> alias name.
     * @return Private key.
     */
    @Override
    public PrivateKey getPrivateKey(final String alias) {

        if (log.isDebugEnabled()) {
            log.debug("getPrivateKey() - begin. Alias=" + this.alias);
        }

        PrivateKey pk = null;

        try {
            char[] cc = this.aliasPwd.toCharArray();
            pk = (PrivateKey)this.ks.getKey(this.alias, cc);
        } catch (Exception e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
        }

        if (log.isDebugEnabled()) {
            if (pk != null) {
                log.debug("getPrivateKey() - end. Ok=true, Algorithm=" + pk.getAlgorithm() + ", Format=" + pk.getFormat());
            } else {
                log.debug("getPrivateKey() - end. Ok=false");
            }
        }

        return pk;
    }

    /**
     * Returns always the selected key alias.
     * </p>
     *
     * @param keyType <code>String</code>
     * @param issuers <code>Principal</code>
     * @param engine  <code>SSLEngine</code> SSl engine.
     * @return Key alias as a String.
     */
    @Override
    public String chooseEngineClientAlias(final String[] keyType, final Principal[] issuers, final SSLEngine engine) {

        if (log.isDebugEnabled()) {
            log.debug("chooseEngineClientAlias() - Alias=" + this.alias);
        }

        return this.alias;
    }

    /**
     * Falls back to the original KeyManager.
     * </p>
     *
     * @param keyType <code>String</code>
     * @param issuers <code>Principal</code>
     * @param engine  <code>SSLEngine</code> SSl engine.
     * @return Key alias as a String.
     */
    @Override
    public String chooseEngineServerAlias(final String keyType, final Principal[] issuers, final SSLEngine engine) {

        if (log.isDebugEnabled()) {
            log.debug("chooseEngineServerAlias() - Alias=" + this.alias);
        }

        return this.alias;
    }
}
