package org.verohallinto.apitamoclient.sec;

import org.apache.commons.lang.exception.ExceptionUtils;
import org.apache.cxf.configuration.jsse.TLSClientParameters;
import org.apache.cxf.configuration.security.FiltersType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.verohallinto.apitamoclient.util.Utils;

import javax.net.ssl.*;
import java.security.KeyStore;
import java.util.Arrays;
import java.util.List;

/**
 * TLS parameters.
 */
public class TLSParameters {

    private static final Logger log = LoggerFactory.getLogger(TLSParameters.class);
    private static final List<String> INCLUDE_CIPHERS = Arrays.asList(".*_WITH_DES_.*",
                                                                      ".*_WITH_AES_.*");
    private static final List<String> EXCLUDE_CIPHERS = Arrays.asList(".*_WITH_DES_.*",
                                                                      ".*_WITH_AES_.*",
                                                                      ".*NULL.*",
                                                                      ".*anon.*",
                                                                      ".*EXPORT.*",
                                                                      ".*MD5",
                                                                      ".*RC4.*");
    /*
        private static final List<String> INCLUDE_CIPHERS = Arrays.asList("TLS_DHE-RSA-AES-128-CBC-SHA256",
                                                                          "TLS_DHE-RSA-AES-256-CBC-SHA256",
                                                                          "TLS_DHE-RSA-AES-128-GCM-SHA256",
                                                                          "TLS_DHE-RSA-AES-256-GCM-SHA384",
                                                                          "TLS_ECDHE-RSA-AES-128-CBC-SHA256",
                                                                          "TLS_ECDHE-RSA-AES-256-CBC-SHA384",
                                                                          "TLS_ECDHE-RSA-AES-128-GCM-SHA256",
                                                                          "TLS_ECDHE-RSA-AES-256-GCM-SHA384",
                                                                          "TLS_ECDHE-ECDSA-AES-128-CBC-SHA256",
                                                                          "TLS_ECDHE-ECDSA-AES-256-CBC-SHA384",
                                                                          "TLS_ECDHE-ECDSA-AES-128-GCM-SHA256",
                                                                          "TLS_ECDHE-ECDSA-AES-256-GCM-SHA384");
        private static final List<String> INCLUDE_CIPHERS = Arrays.asList("TLS_ECDHE-ECDSA-AES256-GCM-SHA384",
                                                                          "TLS_ECDHE-ECDSA-AES128-GCM-SHA256",
                                                                          "TLS_ECDHE-RSA-AES256-GCM-SHA384",
                                                                          "TLS_ECDHE-RSA-AES128-GCM-SHA256",
                                                                          "TLS_ECDHE-ECDSA-AES256-SHA384",
                                                                          "TLS_ECDHE-ECDSA-AES128-SHA256",
                                                                          "TLS_ECDHE-RSA-AES256-SHA384",
                                                                          "TLS_ECDHE-RSA-AES128-SHA256",
                                                                          "TLS_ECDHE-ECDSA-AES256-SHA",
                                                                          "TLS_ECDHE-ECDSA-AES128-SHA",
                                                                          "TLS_ECDHE-RSA-AES256-CBC-SHA",
                                                                          "TLS_ECDHE-RSA-AES128-CBC-SHA");
        private static final List<String> EXCLUDE_CIPHERS = Arrays.asList(".*_EXPORT_.*",
                                                                          ".*_EXPORT1024_.*",
                                                                          ".*_WITH_DES_.*",
                                                                          ".*_WITH_NULL_.*",
                                                                          ".*_DH_anon_.*");
    */
    private String ksJks;
    private String ksJksPwd;
    private String ksCrtAlias;
    private String ksCrtAliasPwd;
    private String tsJks;
    private String tsJksPwd;
    private String acceptHost;

    /**
     * Constructor.
     * </p>
     *
     * @param ksJks         <code>String</code> KeyStore.
     * @param ksJksPwd      <code>String</code> KeyStore password.
     * @param ksCrtAlias    <code>String</code> KeyStore certificate alias.
     * @param ksCrtAliasPwd <code>String</code> KeyStore certificate alias password.
     * @param tsJks         <code>String</code> TrustStore.
     * @param tsJksPwd      <code>String</code> TrustStore password.
     * @param acceptHost    <code>String</code> hostname to accept.
     * @throws IllegalArgumentException Some of the arguments is null or empty.
     */
    public TLSParameters(final String ksJks,
                         final String ksJksPwd,
                         final String ksCrtAlias,
                         final String ksCrtAliasPwd,
                         final String tsJks,
                         final String tsJksPwd,
                         final String acceptHost) {

        if (Utils.isEmpty(ksJks)) {
            throw new IllegalArgumentException("KeyStore cannot be empty.");
        }

        if (Utils.isEmpty(ksJksPwd)) {
            throw new IllegalArgumentException("The KeyStore password cannot be empty.");
        }

        if (Utils.isEmpty(ksCrtAlias)) {
            throw new IllegalArgumentException("The KeyStore alias cannot be empty.");
        }

        if (Utils.isEmpty(ksCrtAliasPwd)) {
            throw new IllegalArgumentException("The KeyStore alias password cannot be empty.");
        }

        if (Utils.isEmpty(tsJks)) {
            throw new IllegalArgumentException("TrustStore cannot be empty.");
        }

        if (Utils.isEmpty(tsJksPwd)) {
            throw new IllegalArgumentException("The TrustStore password cannot be empty.");
        }

        if (Utils.isEmpty(acceptHost)) {
            throw new IllegalArgumentException("The accepted hostname cannot be empty.");
        }

        this.ksJks = ksJks;
        this.ksJksPwd = ksJksPwd;
        this.ksCrtAlias = ksCrtAlias;
        this.ksCrtAliasPwd = ksCrtAliasPwd;
        this.tsJks = tsJks;
        this.tsJksPwd = tsJksPwd;
        this.acceptHost = acceptHost;
    }

    /**
     * Constructs client's TLS parameters.
     * </p>
     *
     * @return TLSClientParameters object.
     */
    public TLSClientParameters getTLSClientParameters() {

        if (log.isDebugEnabled()) {
            log.debug("getTLSClientParameters() - begin.");
        }

        final TLSClientParameters tlsParams = new TLSClientParameters();
        tlsParams.setSecureSocketProtocol("TLS");
        tlsParams.setDisableCNCheck(false);
        tlsParams.setHostnameVerifier(getHostnameVerifier());
        tlsParams.setKeyManagers(getKeyManagers());
        tlsParams.setTrustManagers(getTrustManagers());
        //tlsParams.setCipherSuitesFilter(getCipherSuitesFilter());

        if (log.isDebugEnabled()) {
            getCipherSuitesFilter();
        }

        if (log.isDebugEnabled()) {
            log.debug("getTLSClientParameters() - end.");
        }

        return tlsParams;
    }

    /**
     * Constructs client's TrustManager.
     * </p>
     *
     * @return TrustManagers.
     */
    private TrustManager[] getTrustManagers() {

        if (log.isDebugEnabled()) {
            log.debug("getTrustManagers() - begin.");
        }

        TrustManager[] tm = null;
        final KeyStore keyStore = Utils.getKeyStore(this.tsJks, this.tsJksPwd);
        final String alg = TrustManagerFactory.getDefaultAlgorithm();

        try {
            final TrustManagerFactory tmf = TrustManagerFactory.getInstance(alg);

            if (keyStore != null) {
                tmf.init(keyStore);
            }

            tm = tmf.getTrustManagers();
        } catch (Exception e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
        }

        if (log.isDebugEnabled()) {
            log.debug("getTrustManagers() - end. ok=" + (tm != null));
        }

        return tm;
    }

    /**
     * Constructs client's KeyManager.
     * </p>
     *
     * @return KeyManagers.
     */
    private KeyManager[] getKeyManagers() {

        if (log.isDebugEnabled()) {
            log.debug("getKeyManagers() - begin.");
        }

        KeyManager[] km = null;
        final KeyStore keyStore = Utils.getKeyStore(this.ksJks, this.ksJksPwd);

        if (keyStore != null) {
            try {
                km = new KeyManager[]{new AliasKeyManager(keyStore, this.ksCrtAlias, this.ksCrtAliasPwd)};
            } catch (Exception e) {
                log.error(ExceptionUtils.getFullStackTrace(e));
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("getKeyManagers() - end. ok=" + (km != null));
        }

        return km;
    }

    /**
     * Creates CipherSuitesFilter.
     * </p>
     *
     * @return CipherSuitesFilter object.
     */
    private FiltersType getCipherSuitesFilter() {

        if (log.isDebugEnabled()) {
            log.debug("getCipherSuitesFilter() - begin.");
        }

        final FiltersType filter = new FiltersType();
        //filter.getInclude().addAll(INCLUDE_CIPHERS);
        //filter.getExclude().addAll(EXCLUDE_CIPHERS);

        if (log.isDebugEnabled()) {
            log.debug("getCipherSuitesFilter() - include=" + filter.getInclude());
            log.debug("getCipherSuitesFilter() - exclude=" + filter.getExclude());
            log.debug("getCipherSuitesFilter() - end.");
        }

        return filter;
    }

    /**
     * Creates Hostname Verifier.
     * </p>
     *
     * @return HostnameVerifier object.
     */
    private HostnameVerifier getHostnameVerifier() {

        if (log.isDebugEnabled()) {
            log.debug("getHostnameVerifier() - acceptedHost=" + this.acceptHost);
        }

        return new HostnameVerifier() {
            public boolean verify(final String hostname, final SSLSession session) {
                final boolean ok = hostname != null && hostname.equalsIgnoreCase(acceptHost);

                if (log.isDebugEnabled()) {
                    log.debug("verify() - hostname=" + hostname + ", ok=" + ok + ", CipherSuite=" + session.getCipherSuite());
                }

                return ok;
            }
        };
    }
}
