package org.verohallinto.apitamoclient.util;

import org.apache.commons.lang.exception.ExceptionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.security.KeyStore;
import java.util.HashMap;
import java.util.Map;

/**
 * Static helper methods.
 */
public class Utils {

    private static Logger log = LoggerFactory.getLogger(Utils.class);
    // Environment specific WSDL-files
    private static Map<String, URL> WSDL;

    static {
        WSDL = new HashMap<>();
    }

    private Utils() {
    }

    /**
     * Returns an environment specific WSDL-file.
     * </p>
     *
     * @return URL of the WSDL.
     * @throws FileNotFoundException WSDL was not found.
     */
    public static URL getWsdl() throws FileNotFoundException {

        if (log.isDebugEnabled()) {
            log.debug("getWsdl() - begin. env=" + Props.getEnv());
        }

        URL wsdlFile;
        final String env = Props.getEnv();

        if (WSDL.containsKey(env)) {
            wsdlFile = WSDL.get(env);
        } else {
            final String wsdl = Props.getWsdl();

            if ((wsdlFile = createWsdl(wsdl, env)) != null) {
                WSDL.put(env, wsdlFile);
            } else {
                throw new FileNotFoundException("File not found: " + wsdl);
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("getWsdl() - end. wsdl=" + wsdlFile);
        }

        return wsdlFile;
    }

    /**
     * Creates an environment specific WSDL-file.
     * </p>
     *
     * @param wsdl <code>String</code> original WSDL-file.
     * @param env  <code>String</code> environment.
     * @return Environment specific WSDL-file.
     */
    private static URL createWsdl(final String wsdl, final String env) {

        if (log.isDebugEnabled()) {
            log.debug("createWsdl() - begin. wsdl=" + wsdl + ", env=" + env);
        }

        URL url = null;
        BufferedInputStream bis = null;

        try {
            bis = new BufferedInputStream(Utils.class.getClassLoader().getResourceAsStream(wsdl));
            int x;
            final Document doc = getDocument(bis);

            if (doc != null) {
                final File parent = new File(System.getProperty("java.io.tmpdir"));

                if ((x = wsdl.lastIndexOf('.')) != -1) {
                    String s = wsdl.substring(0, x) + "-" + env + wsdl.substring(x + 1);
                    File file = new File(parent, s);

                    if (!file.getParentFile().exists()) {
                        file.getParentFile().mkdirs();
                    }

                    if (saveWsdl(doc, file)) {
                        try {
                            url = file.toURI().toURL();
                        } catch (MalformedURLException e) {
                            log.error(ExceptionUtils.getFullStackTrace(e));
                        }
                    }
                }
            } else {
                log.error("createWsdl() - Document not found: " + wsdl);
            }

        } finally {
            if (bis != null) {
                try {
                    bis.close();
                } catch (IOException ignore) {
                }
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("createWsdl() - end. wsdl=" + url);
        }

        return url;
    }

    /**
     * Saves an environment specific WSDL-file.
     * </p>
     *
     * @param doc  <code>Document</code> WSDL-document.
     * @param file <code>File</code> file
     * @return true if save was successful else false.
     */
    private static boolean saveWsdl(final Document doc, final File file) {

        if (log.isDebugEnabled()) {
            log.debug("saveWsdl() - begin. file=" + file);
        }

        boolean ok = false;
        BufferedWriter bw = null;

        try {
            bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), StandardCharsets.UTF_8));
            final StringWriter sw = new StringWriter();
            final Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.transform(new DOMSource(doc), new StreamResult(sw));

            bw.write(new String(sw.toString().getBytes(StandardCharsets.UTF_8)));
            bw.flush();
            ok = true;
        } catch (Exception e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
        } finally {
            if (bw != null) {
                try {
                    bw.close();
                } catch (IOException ignore) {
                }
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("saveWsdl() - end. ok=" + ok);
        }

        return ok;
    }

    /**
     * Creates a document from a file.
     * </p>
     *
     * @param bis <code>BufferedInputStream</code> document.
     * @return Document object.
     */
    private static Document getDocument(final BufferedInputStream bis) {

        if (log.isDebugEnabled()) {
            log.debug("getDocument() - begin.");
        }

        Document doc = null;
        boolean ok = false;

        try {
            final DocumentBuilder db = getDocumentBuilder();
            doc = db.parse(bis);
            doc.normalize();

            ok = setLocation(doc);
        } catch (Exception e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
        }

        if (log.isDebugEnabled()) {
            log.debug("getDocument() - end. ok=" + ok);
        }

        return ok ? doc : null;
    }

    /**
     * Creates DocumentBuilder object.
     * </p>
     *
     * @return DocumentBuilder object.
     */
    private static DocumentBuilder getDocumentBuilder() throws ParserConfigurationException {

        final DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        dbf.setValidating(false);
        dbf.setNamespaceAware(true);
        dbf.setIgnoringComments(true);

        return dbf.newDocumentBuilder();
    }

    /**
     * Changes the location attribute.
     * </p>
     * /wsdl:definitions/wsdl:service/wsdl:port/soap:address/@location
     *
     * @param doc <code>Document</code> WSDL-file.
     * @return true = ok, else false.
     */
    private static boolean setLocation(final Document doc) {

        if (log.isDebugEnabled()) {
            log.debug("setLocation() - begin.");
        }

        boolean ok = false;
        final NodeList nl = doc.getElementsByTagNameNS("http://schemas.xmlsoap.org/wsdl/soap/",
                                                       "address");

        if (nl.getLength() == 1) {
            final Element address = (Element)nl.item(0);
            final String location = Props.get("apitamo.url");
            address.setAttribute("location", location);
            ok = true;
        }

        if (!ok) {
            log.error("setLocation() - Error in element '/wsdl:definitions/wsdl:service/wsdl:port/soap:address'");
        }

        if (log.isDebugEnabled()) {
            log.debug("setLocation() - end. ok=" + ok);
        }

        return ok;
    }

    /**
     * Loads KeyStore.
     * </p>
     *
     * @param jks      <code>String</code> KeyStore.
     * @param password <code>String</code> KeyStore password.
     * @return KeyStore or null if KeyStore was not found.
     */
    public static KeyStore getKeyStore(final String jks, final String password) {

        if (log.isDebugEnabled()) {
            log.debug("getKeyStore() - begin. jks=" + jks);
        }

        KeyStore keyStore = null;
        BufferedInputStream bis = null;

        try {
            keyStore = KeyStore.getInstance("JKS");

            if (jks.contains("\\") || jks.contains("/")) {
                bis = new BufferedInputStream(new FileInputStream(jks));
            } else {
                ClassLoader cl = Utils.class.getClassLoader();
                bis = new BufferedInputStream(cl.getResourceAsStream(jks));
            }

            keyStore.load(bis, password != null ? password.toCharArray() : null);
        } catch (Exception e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
        } finally {
            if (bis != null) {
                try {
                    bis.close();
                } catch (IOException ignore) {
                }

                bis = null;
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("getKeyStore() - end. ok=" + (keyStore != null));
        }

        return keyStore;
    }

    /**
     * Checks if given string is null or empty.
     * </p>
     *
     * @param s <code>String</code> a string to check.
     * @return true if string is null or empty.
     */
    public static boolean isEmpty(final String s) {

        return s == null || s.trim().isEmpty();
    }

    /**
     * Checks if given string is numeric.
     * </p>
     *
     * @param s <code>String</code> string to check.
     * @return true if string is numeric else false.
     */
    public static boolean isNumeric(final String s) {

        if (s != null && !s.isEmpty()) {
            final char[] cc = s.toCharArray();

            for (char c : cc) {
                if (c < '0' || c > '9') {
                    return false;
                }
            }

            return true;
        }

        return false;
    }
}
